from h_message_bus import NatsPublisherAdapter
from h_message_bus.domain.request_messages.twitter_get_user_request_message import TwitterGetUserRequestMessage
from h_message_bus.domain.request_messages.twitter_get_user_response_message import TwitterGetUserResponseMessage
from h_message_bus.domain.request_messages.vector_read_metadata_request_message import VectorReadMetaDataRequestMessage
from h_message_bus.domain.request_messages.vector_read_metadata_response_message import \
    VectorReadMetaDataResponseMessage
from h_message_bus.domain.request_messages.vector_save_request_message import VectorSaveRequestMessage

from .priority_queue_service import PriorityQueueService
from ..application.web_docs_service import WebDocsService
from ..infrastructure.priorityqueue.in_memory_priority_queue_repository import InMemoryPriorityQueueRepository


class HaiService:
    def __init__(self, nats_publisher_adapter: NatsPublisherAdapter):
        self.nats_publisher_adapter = nats_publisher_adapter
        self.web_docs_service = WebDocsService()
        queue = InMemoryPriorityQueueRepository()
        self.queue_service = PriorityQueueService(queue)

    async def detect_and_store_documentation(self, twitter_screen_name: str):
        req_message = TwitterGetUserRequestMessage.create_message(twitter_screen_name)
        response = await self.nats_publisher_adapter.request(req_message)
        twitter_user = TwitterGetUserResponseMessage.from_hai_message(response)

        if twitter_user.url is not None:
            print(f"Documentation found for {twitter_user.screen_name}: {twitter_user.url}")
            docs = await self.web_docs_service.discover_documentation(twitter_user.url)

            for doc in docs:
                collection_name = f"{twitter_user.screen_name}_docs"
                chapters = doc.chapters
                for chapter in chapters:
                    i = 0
                    for text in chapter.paragraphs:
                        document_id = f"{doc.title}_{chapter.heading}_{i}"

                        req_metadata = {
                            "source": doc.url
                        }
                        i = i + 1

                        request = VectorSaveRequestMessage.create_message(
                            collection_name=collection_name,
                            document_id=document_id,
                            content=text,
                            metadata=req_metadata)

                        await self.nats_publisher_adapter.publish(request)

        else:
            print(f"No documentation found for {twitter_user.screen_name}")

    async def load_current_knowledge_base_metadata(self):
        message = VectorReadMetaDataRequestMessage.create_message()
        response = await self.nats_publisher_adapter.request(message)
        metadata_result = VectorReadMetaDataResponseMessage.from_hai_message(response)
        print(metadata_result)



class EcosystemPatternRepository:
    """Repository of patterns that identify ecosystem relationships, builders, and collaboration"""

    def __init__(self):
        # Domains commonly associated with ecosystem and project showcases
        self.ecosystem_domains = [
            "showcase.",
            "ecosystem.",
            "community.",
            "gallery.",
            "partners.",
            "developers.",
            "marketplace.",
            "expo.",
            "apps.",
            "extensions.",
            "plugins.",
        ]

        # URL path patterns indicating ecosystem/builder content
        self.ecosystem_path_patterns = [
            r"/ecosystem/",
            r"/showcase/",
            r"/community/",
            r"/built-with/",
            r"/case-studies/",
            r"/customers/",
            r"/partners/",
            r"/users/",
            r"/success-stories/",
            r"/integrations/",
            r"/extensions/",
            r"/marketplace/",
            r"/plugins/",
            r"/addons/",
            r"/gallery/",
            r"/examples/",
            r"/projects/",
            r"/contributors/",
            r"/whos-using/",
        ]

        # Link text patterns suggesting ecosystem content
        self.ecosystem_text_patterns = [
            r"(?i)ecosystem",
            r"(?i)showcase",
            r"(?i)built with",
            r"(?i)powered by",
            r"(?i)case stud(y|ies)",
            r"(?i)success stor(y|ies)",
            r"(?i)who('s| is) using",
            r"(?i)our users",
            r"(?i)our customers",
            r"(?i)integrations?",
            r"(?i)extensions?",
            r"(?i)plugins?",
            r"(?i)addons?",
            r"(?i)community projects",
            r"(?i)community contributions",
            r"(?i)user contributions",
            r"(?i)featured projects",
            r"(?i)gallery",
        ]

        # Header/title patterns suggesting ecosystem sections
        self.ecosystem_header_patterns = [
            r"(?i)ecosystem",
            r"(?i)who('s| is) using",
            r"(?i)built (on|with)",
            r"(?i)powered by",
            r"(?i)trusted by",
            r"(?i)customer(s| success)",
            r"(?i)case stud(y|ies)",
            r"(?i)success stor(y|ies)",
            r"(?i)showcase",
            r"(?i)featured (users|customers|projects)",
            r"(?i)community (projects|showcase)",
            r"(?i)partner(s| program)",
            r"(?i)(our|notable) users",
            r"(?i)companies using",
            r"(?i)in production",
            r"(?i)contributor(s| showcase)",
            r"(?i)extension (gallery|showcase)",
            r"(?i)plugin (directory|marketplace)",
            r"(?i)apps? (built|marketplace|gallery)",
        ]

        # Content phrases that suggest ecosystem descriptions
        self.ecosystem_content_patterns = [
            r"(?i)built (on|with) (our|this)",
            r"(?i)(companies|organizations|projects) (using|powered by)",
            r"(?i)(is|are) using (our|this)",
            r"(?i)powered by (our|this)",
            r"(?i)extend(s|ing)? (the|our) (platform|ecosystem)",
            r"(?i)integrated with",
            r"(?i)build(s|ing)? (on top of|with)",
            r"(?i)leverage(s|ing)? (our|this)",
            r"(?i)extend(s|ing)? (the|our) (functionality|capabilities)",
            r"(?i)based on (our|this)",
            r"(?i)implemented (with|using)",
            r"(?i)developed (with|using)",
            r"(?i)(join|be part of) (our|the) ecosystem",
        ]

        # Builder and contribution-specific patterns
        self.builder_patterns = [
            r"(?i)how to (build|contribute)",
            r"(?i)build(ing)? (with|on)",
            r"(?i)develop(ing)? (with|on)",
            r"(?i)contribute to",
            r"(?i)contributor guide",
            r"(?i)developer program",
            r"(?i)join (our|the) (ecosystem|community)",
            r"(?i)become a (contributor|partner)",
            r"(?i)extend (our|the) (platform|ecosystem)",
            r"(?i)create (your own|an?) (plugin|extension|integration)",
            r"(?i)developer (resources|portal)",
            r"(?i)sdk",
            r"(?i)api (access|integration)",
            r"(?i)partner (program|portal)",
        ]

        # Visual cues that often indicate ecosystem showcases
        self.visual_indicators = [
            r"logo grid",
            r"logo carousel",
            r"client logos",
            r"partner logos",
            r"customer logos",
            r"company logos",
            r"card gallery",
            r"project cards",
            r"showcase gallery",
            r"case study cards",
            r"testimonials",
            r"user testimonials",
        ]

        # Collaboration-specific patterns
        self.collaboration_patterns = [
            r"(?i)how to collaborate",
            r"(?i)collaboration (guide|opportunities)",
            r"(?i)working together",
            r"(?i)partner(ship|ing) (opportunities|program)",
            r"(?i)join (our|the) (community|ecosystem)",
            r"(?i)community (contribution|participation)",
            r"(?i)open (source|collaboration)",
            r"(?i)contribute (code|documentation|resources)",
            r"(?i)become a (partner|contributor|maintainer)",
            r"(?i)collaboration (framework|model)",
            r"(?i)(business|technical) partnership",
            r"(?i)developer relations",
            r"(?i)community (engagement|involvement)",
        ]

        # Key meta tags that might indicate ecosystem content
        self.meta_tag_patterns = [
            r"(?i)ecosystem",
            r"(?i)showcase",
            r"(?i)community",
            r"(?i)partner program",
            r"(?i)integration",
            r"(?i)extension",
            r"(?i)plugin",
            r"(?i)marketplace",
            r"(?i)collaboration",
            r"(?i)use cases",
            r"(?i)case studies",
            r"(?i)success stories",
        ]

        # Schema.org types that often indicate ecosystem relationships
        self.schema_types = [
            "Product",
            "SoftwareApplication",
            "Organization",
            "BusinessPartner",
            "ProgramMembership",
            "CreativeWork",
            "SoftwareSourceCode",
            "WebApplication",
        ]


import hashlib
from dataclasses import dataclass, field
from typing import Dict, List

from ....domain.web_docs.gitbook.text_chapter import TextChapter


@dataclass
class TextPage:
    """Represents text on a page from a web document"""
    url: str = ""
    title: str = ""
    content: str = ""
    last_updated: str = ""

    index: int = 0
    toc_level: int = 0
    parent_id: str = ""

    chapters: List[TextChapter] = field(default_factory=list)
    links: Dict[str, str] = field(default_factory=dict)  # Text -> URL

    id: str = field(init=False, default="")
    content_hash: str = field(init=False, default="")

    def __post_init__(self):
        self.id = hashlib.md5(self.url.encode()).hexdigest()
        # Generate a content hash for deduplication
        self.content_hash = hashlib.md5(self.content.encode()).hexdigest() if self.content else ""

    def to_dict(self):
        """Convert this TextPage instance to a serializable dictionary"""
        result = {
            'url': self.url,
            'title': self.title,
            #'content': self.content,
            'last_updated': self.last_updated,
            'index': self.index,
            'toc_level': self.toc_level,
            'parent_id': self.parent_id,
            'id': self.id,
            'content_hash': self.content_hash,
            'links': self.links,
            'chapters': [chapter.to_dict() for chapter in self.chapters]
        }
        return result



import uuid

import requests

from ....domain.reasoning.llm_generate_respository import LlmGenerateRepository


class OllamaGenerateRepository(LlmGenerateRepository):

    def __init__(self, api_url: str, model_name: str, system_prompt: str = None, temperature: float = None, seed: int = None):
        self.model_name = model_name
        self.system_prompt = system_prompt
        self.api_url = api_url
        self.temperature = temperature
        self.seed = seed


    def generate(self, user_prompt: str, system_prompt: str = None, session_id: str = None) -> str|None:
        url = f"{self.api_url}/generate"
        random_guid = uuid.uuid4()
        guid_str = str(random_guid)
        system_prompt = system_prompt or self.system_prompt
        payload = {
            "model": self.model_name,
            "prompt": user_prompt,
            "system": system_prompt,
            "stream": False,
            "session": guid_str,
            "num_ctx": "5000",
            "temperature": "0.6"
        }

        if session_id:
            payload["session"] = session_id
        if self.seed:
            payload["seed"] = self.seed
        if self.temperature:
            payload["temperature"] = self.temperature

        try:
            print(payload)
            response = requests.post(url, json=payload)
            response.raise_for_status()

            print(response.json())

            response_content = response.json()["response"]
            return clean_llm_response(response_content)

        except requests.exceptions.RequestException as e:
            print(f"Error occurred during API call: {e}")
            return None




import queue
import threading
from queue import PriorityQueue
from typing import Dict, List, Optional

from ...domain.priorityqueue.priority_queue_repository import PriorityQueueRepository
from ...domain.priorityqueue.queue_item import QueueItem


class InMemoryPriorityQueueRepository(PriorityQueueRepository):
    """In-memory implementation of the PriorityQueueRepository using Python's PriorityQueue"""

    def __init__(self):
        # Dictionary mapping queue names to their PriorityQueue instances
        self.queues: Dict[str, PriorityQueue] = {}
        # Locks to ensure thread safety
        self.locks: Dict[str, threading.Lock] = {}

    def _get_or_create_queue(self, queue_name: str, maxsize: int = 10000) -> PriorityQueue:
        """Get or create a queue with the given name"""
        if queue_name not in self.queues:
            self.queues[queue_name] = PriorityQueue(maxsize=maxsize)
            self.locks[queue_name] = threading.Lock()
        return self.queues[queue_name]

    def _get_lock(self, queue_name: str) -> threading.Lock:
        """Get the lock for the specified queue"""
        if queue_name not in self.locks:
            self.locks[queue_name] = threading.Lock()
        return self.locks[queue_name]

    def add_item(self, queue_name: str, item: QueueItem) -> None:
        """Add an item to the specified queue"""
        q = self._get_or_create_queue(queue_name)
        with self._get_lock(queue_name):
            # The queue automatically orders by priority
            q.put(item)

    def get_highest_priority_item(self, queue_name: str, block: bool = False, timeout: Optional[float] = None) -> Optional[QueueItem]:
        """Get and remove the highest priority item from the queue"""
        if queue_name not in self.queues:
            return None

        q = self.queues[queue_name]
        try:
            # We need to acquire the lock to check initially, but then we need to release it
            # so that other threads can add items while we're blocking
            with self._get_lock(queue_name):
                # If not blocking and queue is empty, return immediately
                if not block and q.empty():
                    return None

            # Note: queue.get() with block=True will internally manage locking
            # We don't hold our custom lock during the blocking wait
            return q.get(block=block, timeout=timeout)
        except queue.Empty:
            # This will be raised if timeout occurs with no items
            return None


    def get_items(self, queue_name: str, limit: int = 10) -> List[QueueItem]:
         """Get multiple items from the queue in priority order without removing them"""
         if queue_name not in self.queues:
             return []

         q = self.queues[queue_name]
         result = []

         with self._get_lock(queue_name):
             # Create a temporary list to hold items that we'll put back
             temp_items = []

             # Get up to 'limit' items
             count = 0
             while not q.empty() and count < limit:
                 item = q.get()
                 temp_items.append(item)
                 result.append(item)
                 count += 1

             # Put all the items back in the same order
             for item in temp_items:
                 q.put(item)

         return result

    def queue_length(self, queue_name: str) -> int:
        """Get the number of items in the queue"""
        if queue_name not in self.queues:
            return 0

        q = self.queues[queue_name]
        with self._get_lock(queue_name):
            return q.qsize()

    def get_queue_names(self) -> List[str]:
        """Get a list of all available queue names"""
        return list(self.queues.keys())



import logging
import re
from typing import List, Optional
from urllib.parse import urljoin

from bs4 import BeautifulSoup, Tag

from ...domain.web_docs.gitbook.text_chapter import TextChapter
from ...domain.web_docs.web_link import WebLink

logger = logging.getLogger(__name__)

class SoupProcessor:
    def __init__(self, html_content:str):
        self.soup = BeautifulSoup(html_content, 'html.parser')

    def extract_links(self, base_url: str) -> List[WebLink]:
        """Extract links from a page"""
        web_links = []
        links = self.soup.find_all('a', href=True)
        for link in links:
            href = link.get('href', '').strip()
            if not href or href.startswith('#') or href.startswith('javascript:'):
                continue

            full_url = urljoin(base_url, href)
            link_text = link.get_text().strip()
            web_link = WebLink(url=full_url, title=link_text)
            web_links.append(web_link)
        return web_links

    def normalize_url(self, href, base_url) -> Optional[str]:
        """Normalize URL to absolute form and filter out non-content URLs"""
        # Skip fragment-only URLs
        if href.startswith('#'):
            return None

        # Skip external links
        if href.startswith(('http://', 'https://')) and not href.startswith(base_url):
            return None

        # Skip resource URLs
        if href.endswith(('.jpg', '.jpeg', '.png', '.gif', '.svg', '.pdf', '.zip', '.js', '.css')):
            return None

        # Convert to absolute URL if needed
        full_url = href
        if not href.startswith(('http://', 'https://')):
            full_url = urljoin(base_url, href)

        # Make sure URL belongs to the same domain
        if not full_url.startswith(base_url):
            return None

        return full_url

    def extract_last_updated_refs_from_soup(self) -> str:
        datetime_value = ""

        # Find and remove elements containing "Last updated" text
        for element in self.soup.find_all(string=lambda text: text and "Last updated" in text):
            # Get the parent element and remove it
            parent = element.parent
            if parent:
                parent.decompose()

        return datetime_value

    def extract_title(self) -> Optional[str]:
        """Extract the title of the page using multiple strategies"""
        # Strategy 1: Look for h1
        h1 = self.soup.find('h1')
        if h1:
            return h1.get_text(strip=True)

        # Strategy 2: Look for title tag
        title_tag = self.soup.find('title')
        if title_tag:
            title_text = title_tag.get_text(strip=True)
            title_parts = re.split(r'[|\-–]', title_text)
            return title_parts[0].strip()

        # Strategy 3: Try to find GitBook-specific title elements
        gitbook_title = self.soup.find('span', {'data-testid': 'page.title'})
        if gitbook_title:
            return gitbook_title.get_text(strip=True)

        return None

    def find_body_content(self) -> Optional[Tag]:
        body_content = self.soup.find('body')
        if body_content:
            return body_content
        return None

    def gitbook_extract_modern_nav(self, base_url, processed_urls):
        """Extract navigation from modern GitBook layout"""
        nav_links = []

        # Look for navigation sidebar
        sidebar = self.soup.select_one('div[data-testid="page.desktopTableOfContents"]')
        if sidebar:
            for link in sidebar.find_all('a', href=True):
                full_url = self.normalize_url(link['href'], base_url)
                if full_url and full_url not in processed_urls:
                    nav_links.append(full_url)
                    processed_urls.add(full_url)

        return nav_links

    def gitbook_extract_traditional_nav(self, base_url, processed_urls):
        """Extract navigation from traditional GitBook layout"""
        nav_links = []

        # Find GitBook navigation elements
        nav_elements = self.soup.find_all(['nav', 'aside'])
        for nav in nav_elements:
            # Look for lists that typically contain the navigation
            nav_lists = nav.find_all(['ol', 'ul'])
            for nav_list in nav_lists:
                for li in nav_list.find_all('li'):
                    link = li.find('a', href=True)
                    if link:
                        full_url = self.normalize_url(link['href'], base_url)
                        if full_url and full_url not in processed_urls:
                            nav_links.append(full_url)
                            processed_urls.add(full_url)

        # Try summary element which is common in GitBook
        summary = self.soup.find('ul', {'class': 'summary'})
        if summary:
            for link in summary.find_all('a', href=True):
                full_url = self.normalize_url(link['href'], base_url)
                if full_url and full_url not in processed_urls:
                    nav_links.append(full_url)
                    processed_urls.add(full_url)

        return nav_links

    def gitbook_extract_pagination_links(self, base_url, processed_urls):
        """Extract navigation from pagination elements"""
        nav_links = []

        # Find pagination links (next/prev)
        selectors = [
            'a[aria-label="Next"]',
            'a[aria-label="Previous"]',
            'a.navigation-next',
            'a.navigation-prev',
            'a:has(svg[data-icon="arrow-right"])',
            'a:has(svg[data-icon="arrow-left"])'
        ]

        for selector in selectors:
            try:
                for link in self.soup.select(selector):
                    if link.has_attr('href'):
                        full_url = self.normalize_url(link['href'], base_url)
                        if full_url and full_url not in processed_urls:
                            nav_links.append(full_url)
                            processed_urls.add(full_url)
            except Exception:
                continue

        return nav_links

    def gitbook_extract_class_based_nav(self, base_url, processed_urls):
        """Extract navigation based on common GitBook class patterns"""
        nav_links = []

        # Common class patterns for navigation in GitBook
        class_patterns = [
            'nav-', 'menu-', 'sidebar-', 'toc-', '-nav', '-menu', '-sidebar', '-toc'
        ]

        # Look for elements with these class patterns
        for pattern in class_patterns:
            elements = self.soup.find_all(class_=lambda c: c and pattern in c)
            for element in elements:
                for link in element.find_all('a', href=True):
                    full_url = self.normalize_url(link['href'], base_url)
                    if full_url and full_url not in processed_urls:
                        nav_links.append(full_url)
                        processed_urls.add(full_url)

        return nav_links

    @staticmethod
    def clean_template_usage(content: Tag):
        if not content or not isinstance(content, Tag):
            return None
        # Step 1: Build a mapping of template IDs to hidden div content
        template_map = {}
        # Find all hidden divs with IDs like S:*
        for hidden_div in content.find_all('div', {'hidden': True}, id=re.compile(r'S:\d+')):
            div_id = hidden_div.get('id')
            # Store the first child (e.g., <a> tag) or the entire content
            if hidden_div.contents:
                template_map[div_id] = hidden_div.contents[0] if len(hidden_div.contents) == 1 else hidden_div

        # Step 2: Replace <template> tags with content from hidden divs based on $RS logic
        for template in content.find_all('template', id=re.compile(r'P:\d+')):
            template_id = template.get('id')  # e.g., P:2
            # Convert P:* to S:* to match the hidden div (assuming $RS("S:2", "P:2") pattern)
            source_id = f"S:{template_id.split(':')[1]}"  # e.g., S:2
            if source_id in template_map:
                # Replace the template with the content from the hidden div
                replacement = template_map[source_id]
                # If it's a Tag, use it directly; if it's a div, extract its contents
                if isinstance(replacement, Tag):
                    template.replace_with(replacement)
                else:
                    template.replace_with(replacement.contents[0])

    @staticmethod
    def extract_chapters(content: Tag) -> List[TextChapter]:
        chapters = []

        # Create a default chapter for content before any heading
        default_chapter = TextChapter(heading="Introduction", level=0)
        current_chapter = default_chapter
        chapters.append(default_chapter)

        for element in content.find_all(['h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p']):
            if element.name.startswith('h'):
                # Extract heading level (h1=1, h2=2, etc.)
                level = int(element.name[1])
                heading_text = element.get_text(strip=True)

                # Create a new chapter
                current_chapter = TextChapter(heading=heading_text, level=level)
                chapters.append(current_chapter)

            elif element.name == 'p' and current_chapter is not None:
                paragraph_text = element.get_text(strip=True)
                if paragraph_text:
                    current_chapter.paragraphs.append(paragraph_text)

        # Remove any chapters without content if they're not top-level
        return [ch for ch in chapters if ch.paragraphs or ch.level <= 2]


import logging
from typing import Optional

from playwright.async_api import async_playwright

from ...domain.webpages.web_fetcher_repository import WebFetcherRepository

logger = logging.getLogger(__name__)

class PlayWrightWebContentFetcher(WebFetcherRepository):
    def __init__(self):
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.5'
        }
        self.page_load_timeout = 60
        self.wait_for_idle = True

    async def fetch(self, url: str) -> Optional[str]:
        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=True, args=[
            '--disable-dev-shm-usage',  # Required for Docker
            '--no-sandbox',  # Required for Docker non-root user
            '--disable-setuid-sandbox',  # Required for Docker security
            '--disable-gpu',  # Reduces resource usage
            ])

            logger.debug(
                f"Launching headless browser with user agent: {self.headers.get('User-Agent')}"
            )
            try:
                context = await browser.new_context(
                    user_agent=self.headers.get('User-Agent')
                )
                page = await context.new_page()

                # Set timeout
                page.set_default_timeout(self.page_load_timeout * 1000)  # Convert to ms

                page.on("console", lambda msg: logger.debug(f"Browser console {url}: {msg.text}"))

                # Navigate to the URL
                await page.goto(url, timeout=self.page_load_timeout * 1000, wait_until='domcontentloaded')

                # Wait for network to be idle if requested
                if self.wait_for_idle:
                    logger.debug(f"Waiting for network idle on {url}")

                    await page.wait_for_load_state("networkidle", timeout=self.page_load_timeout * 1000)
                    logger.debug(
                        f"Network idle on {url} after {self.page_load_timeout} seconds"
                    )

                logger.debug(f"Successfully fetched {url} with headless browser")

                # Get the rendered HTML
                return await page.content()

            except Exception as e:
                logger.error(f"Error fetching {url} with headless browser: {str(e)}")
                return None
            finally:
                await browser.close()



import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from h_message_bus import HaiMessage


@dataclass(frozen=True)
class QueueItem:
    """Value object representing an item in the priority queue"""
    id: str
    content: HaiMessage
    priority: int
    created_at: datetime
    metadata: Optional[dict] = None

    @classmethod
    def create(cls, content: HaiMessage, priority: int, metadata: Optional[dict] = None) -> "QueueItem":
        """Factory method to create a new QueueItem"""
        return cls(
            id=str(uuid.uuid4()),
            content=content,
            priority=priority,
            created_at=datetime.utcnow(),
            metadata=metadata
        )

    def __lt__(self, other):
        """Comparison method for priority queue ordering
        - Primary sort by priority: Higher number = higher priority
        - Secondary sort by timestamp: Earlier timestamp = higher priority (FIFO)
        """
        if not isinstance(other, QueueItem):
            return NotImplemented

        # First, compare by priority (higher priority value comes first)
        if self.priority != other.priority:
            return self.priority > other.priority

        # If priorities are equal, compare by timestamp (older timestamp comes first)
        return self.created_at < other.created_at



from typing import List, Optional

from ..domain.web_docs.doc_link_scorer_service import DocLinkScorerService
from ..domain.web_docs.documentation_pattern_repository import DocumentationPatternRepository
from ..domain.web_docs.gitbook.text_page import TextPage
from ..domain.web_docs.gitbook_web_fetcher_service import GitbookWebFetcherService
from ..domain.web_docs.web_docs_link_detector import WebDocsLinkDetector
from ..domain.web_docs.web_link import WebLink
from ..infrastructure.playwright.playwright_web_content_fetcher import PlayWrightWebContentFetcher


class WebDocsService:

    def __init__(self):
        self.pattern_repo = DocumentationPatternRepository()
        self.scorer = DocLinkScorerService(self.pattern_repo)
        self.headless_browser = PlayWrightWebContentFetcher()
        self.web_link_detector = WebDocsLinkDetector(
            self.scorer,
            self.headless_browser)


    async def discover_documentation(self, website_url: str) -> Optional[List[TextPage]]:
        detected_links = await self.detect_documentation_links(website_url)
        for link in detected_links:
            print(f"Fetching {link.url}")
            gitbook_fetcher = GitbookWebFetcherService(link.url)
            gitbook_pages = await gitbook_fetcher.fetch()
            return gitbook_pages

    async def detect_documentation_links(self, website_url: str) -> List[WebLink]:
        """
        Function to detect documentation links from a website
        Returns a list of potential documentation root URLs
        """
        return await self.web_link_detector.find_docs_links(website_url)



import uuid
from typing import List, Dict, Union, Optional
from dataclasses import dataclass


@dataclass
class Sentiment:
    """Domain object representing the sentiment of analyzed text"""
    score: float
    label: str  # 'negative', 'neutral', or 'positive'


@dataclass
class DetectedEntity:
    """Domain object representing an entity detected in analyzed text"""
    name: str
    type: str
    mentions: List[str] = None

    def __post_init__(self):
        if self.mentions is None:
            self.mentions = []


@dataclass
class DetectedRelationship:
    """Domain object representing a relationship between entities detected in text"""
    source_entity_name: str
    relation_type: str
    target_entity_name: str
    confidence: float = 0.0  # Confidence score of this relationship detection


@dataclass
class ExtractedFact:
    """Domain object representing a factual statement extracted from text"""
    statement: str
    referenced_entities: List[str] = None  # Names of entities referenced in this fact

    def __post_init__(self):
        if self.referenced_entities is None:
            self.referenced_entities = []


@dataclass
class TextAnalysisResult:
    """Root aggregate for the result of analyzing a piece of text"""
    text_id: str
    category: str
    topic: str
    sentiment: Sentiment
    entities: List[DetectedEntity]
    relationships: List[DetectedRelationship]
    facts: List[ExtractedFact]

    def __init__(self, text_id: str, category: str, topic: str,
                 sentiment_score: float = 0.0,
                 sentiment_label: str = "neutral",
                 entities: Optional[List[Dict[str, Union[str, List[str]]]]] = None,
                 relationships: Optional[List[Dict[str, Union[str, float]]]] = None,
                 facts: Optional[List[Dict[str, Union[str, List[str]]]]] = None):

        self.text_id = text_id
        self.category = category
        self.topic = topic
        self.sentiment = Sentiment(sentiment_score, sentiment_label)

        # Process entities
        self.entities = []
        if entities:
            for entity_data in entities:
                entity = DetectedEntity(
                    name=entity_data["name"],
                    type=entity_data["type"],
                    mentions=entity_data.get("mentions", [])
                )
                self.entities.append(entity)

        # Process relationships
        self.relationships = []
        if relationships:
            for rel_data in relationships:
                rel = DetectedRelationship(
                    source_entity_name=rel_data["source_entity_name"],
                    relation_type=rel_data["relation_type"],
                    target_entity_name=rel_data["target_entity_name"],
                    confidence=rel_data.get("confidence", 0.0)
                )
                self.relationships.append(rel)

        # Process facts
        self.facts = []
        if facts:
            for fact_data in facts:
                fact = ExtractedFact(
                    statement=fact_data["statement"],
                    referenced_entities=fact_data.get("referenced_entities", [])
                )
                self.facts.append(fact)

    def to_dict(self) -> Dict[str, Union[str, Dict, List]]:
        """Convert the TextAnalysisResult to a dictionary"""
        return {
            "text_id": self.text_id,
            "category": self.category,
            "topic": self.topic,
            "sentiment": {
                "score": self.sentiment.score,
                "label": self.sentiment.label
            },
            "entities": [
                {
                    "name": entity.name,
                    "type": entity.type,
                    "mentions": entity.mentions
                } for entity in self.entities
            ],
            "relationships": [
                {
                    "source_entity_name": rel.source_entity_name,
                    "relation_type": rel.relation_type,
                    "target_entity_name": rel.target_entity_name,
                    "confidence": rel.confidence
                } for rel in self.relationships
            ],
            "facts": [
                {
                    "statement": fact.statement,
                    "referenced_entities": fact.referenced_entities
                } for fact in self.facts
            ]
        }

    @classmethod
    def from_dict(cls, data: Dict) -> 'TextAnalysisResult':
        """Create a TextAnalysisResult from a dictionary"""
        random_guid = uuid.uuid4()
        random_guid_str = str(uuid.uuid4())

        return cls(
            text_id=random_guid_str,
            category=data["category"],
            topic=data["topic"],
            sentiment_score=data["sentiment"]["score"],
            sentiment_label=data["sentiment"]["label"],
            entities=data.get("entities", []),
            relationships=data.get("relationships", []),
            facts=data.get("facts", [])
        )



class DocumentationPatternRepository:
    """Repository of patterns that indicate documentation links"""

    def __init__(self):
        # Domain patterns that commonly host documentation
        self.doc_domains = [
            "gitbook.io",
            "readthedocs.io",
            "docs.github.com",
            "developer.mozilla.org",
            "confluence.",
            "zendesk.com",
            "help.",
            "support.",
            "wiki.",
        ]

        # URL path patterns that commonly indicate documentation
        self.doc_path_patterns = [
            r"/docs/",
            r"/documentation/",
            r"/guide/",
            r"/manual/",
            r"/help/",
            r"/knowledge/",
            r"/support/",
            r"/api/",
            r"/reference/",
            r"/wiki/",
        ]

        # Link text patterns that suggest documentation
        self.doc_text_patterns = [
            r"(?i)documentation",
            r"(?i)docs",
            r"(?i)developer guide",
            r"(?i)user guide",
            r"(?i)knowledge base",
            r"(?i)help center",
            r"(?i)manual",
            r"(?i)api reference",
            r"(?i)getting started",
            r"(?i)learn more",
        ]


def replace_placeholders(prompt_string, placeholder_dict):
    result = prompt_string

    for placeholder, replacement in placeholder_dict.items():
        result = result.replace(placeholder, str(replacement))

    return result



from abc import ABC, abstractmethod
from typing import Optional


class WebFetcherRepository(ABC):

    @abstractmethod
    async def fetch(self, url: str) -> Optional[str]:
        """Fetch the content of the given URL."""
        pass


from typing import List

from ...domain.webpages.web_fetcher_repository import WebFetcherRepository
from ...domain.web_docs.doc_link_scorer_service import DocLinkScorerService
from ...domain.web_docs.web_link import WebLink
from ...infrastructure.beautifulsoup.soup_processor import SoupProcessor


class WebDocsLinkDetector:
    def __init__(self, doc_link_scorer: DocLinkScorerService, web_fetcher: WebFetcherRepository, confidence_threshold: float = 0.5):
        self.doc_link_scorer = doc_link_scorer
        self.web_fetcher = web_fetcher
        self.confidence_threshold = confidence_threshold

    async def find_docs_links(self, website_url: str) -> List[WebLink]:
        doc_links = []

        web_content = await self.web_fetcher.fetch(website_url)
        if not web_content:
            return doc_links

        soup_processor = SoupProcessor(web_content)
        web_links = soup_processor.extract_links(website_url)
        for web_link in web_links:
            score = self.doc_link_scorer.score(web_link.url, web_link.title)
            if score >= self.confidence_threshold:
                doc_links.append(web_link)
        return doc_links


class WebLink:
    def __init__(self, url: str, title: str):
        self.url = url
        self.title = title

    def to_dict(self):
        """Convert this WebLink instance to a serializable dictionary"""
        return {
            'url': self.url,
            'title': self.title
        }



from typing import Dict


class OllamaChatMessage:
    def __init__(self, role: str, content: str):
        self.role = role
        self.content = content

    def to_dict(self) -> Dict[str, str]:
        return {
            "role": self.role,
            "content": self.content
        }




from abc import ABC, abstractmethod
from typing import List, Optional

from .queue_item import QueueItem


class PriorityQueueRepository(ABC):
    """Repository interface for priority queue operations"""

    @abstractmethod
    def add_item(self, queue_name: str, item: QueueItem) -> None:
        """Add an item to the specified queue"""
        pass

    @abstractmethod
    def get_highest_priority_item(self, queue_name: str, block: bool = False, timeout: Optional[float] = None) -> Optional[QueueItem]:
        """Get and remove the highest priority item from the queue"""
        pass

    # @abstractmethod
    # def get_items(self, queue_name: str, limit: int = 10) -> List[QueueItem]:
    #     """Get multiple items from the queue in priority order without removing them"""
    #     pass

    @abstractmethod
    def queue_length(self, queue_name: str) -> int:
        """Get the number of items in the queue"""
        pass

    @abstractmethod
    def get_queue_names(self) -> List[str]:
        """Get a list of all available queue names"""
        pass



from typing import Optional, List

import requests

from ....domain.reasoning.llm_chat_repository import LlmChatRepository
from ....infrastructure.llm.llm_response_cleaner import clean_llm_response
from ....infrastructure.llm.ollama.models.ollama_chat_message import OllamaChatMessage
from ....infrastructure.llm.ollama.models.ollama_chat_session import OllamaChatSession


class OllamaChatRepository(LlmChatRepository):

    def __init__(self, api_url: str, model_name: str, system_prompts: list[str] = None, temperature: float = None, seed: int = None):
        self.api_url = api_url
        self.model_name = model_name
        self.temperature = temperature
        self.seed = seed
        self.system_prompts = system_prompts

    def chat(self, user_message: str, session_id: str) -> Optional[str]:

        messages = [OllamaChatMessage("user", user_message)]
        for system_prompt in self.system_prompts:
            messages.append(OllamaChatMessage("system", system_prompt))
        session = OllamaChatSession(session_id, messages)

        return self._call_ollama_api(session.messages)

    def _call_ollama_api(self, messages: List[OllamaChatMessage]) -> Optional[str]:
        url = f"{self.api_url}/chat"
        formatted_messages = [message.to_dict() for message in messages]
        payload = {
            "model": self.model_name,
            "messages": formatted_messages,
            "stream": False,
            "temperature": "0.6"
        }
        if self.temperature:
            payload["temperature"] = self.temperature
        if self.seed:
            payload["seed"] = self.seed

        try:
            print(payload)
            response = requests.post(url, json=payload)
            response.raise_for_status()
            full_response = response.json()["message"]["content"]
            print(full_response)
            return clean_llm_response(full_response)

        except requests.exceptions.RequestException as e:
            print(f"Error occurred during API call: {e}")
            return None






from abc import ABC, abstractmethod
from typing import List, Optional

from ...domain.web_docs.gitbook.text_page import TextPage


class WebTextFetcherRepository(ABC):

    @abstractmethod
    async def fetch(self) -> Optional[List[TextPage]]:
        """Fetch all content"""
        pass


from dataclasses import dataclass, field
from typing import List


@dataclass
class TextChapter:
    """Represents a chapter/section in a page defined by a heading."""
    heading: str
    level: int  # h1=1, h2=2, etc.
    paragraphs: List[str] = field(default_factory=list)

    def to_dict(self):
        """Convert this TextChapter instance to a serializable dictionary"""
        return {
            'heading': self.heading,
            'level': self.level,
            'paragraphs': self.paragraphs
        }



import asyncio
from typing import Optional, Set, Dict, List
from urllib.parse import urlparse

import aiohttp

from ...domain.web_docs.gitbook.text_page import TextPage
from ...domain.webpages.web_text_fetcher_repository import WebTextFetcherRepository
from ...infrastructure.beautifulsoup.soup_processor import SoupProcessor


class GitbookWebFetcherService(WebTextFetcherRepository):

    def __init__(self, url: str):
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.5'
        }
        self.base_url = url.rstrip('/')
        self.base_domain = urlparse(self.base_url).netloc

        self.delay = 1.0  # Delay between requests in seconds
        self.max_retries = 3
        self.retry_delay = 2  # Initial retry delay in seconds
        self.concurrent_requests = 3  # Number of concurrent requests

        self.visited_urls: Set[str] = set()
        self.content_hashes: Dict[str, str] = {}  # Hash -> URL
        self.pages: Dict[str, TextPage] = {}

    async def fetch(self) -> Optional[List[TextPage]]:
        timeout = aiohttp.ClientTimeout(total=180)  # 3 minutes total timeout
        async with aiohttp.ClientSession(headers=self.headers, timeout=timeout) as session:
            # Start with main page

            await self._process_url(session, self.base_url)

            # Wait for all tasks to complete
            await asyncio.sleep(0)

            # Sort pages by index
            sorted_pages = sorted(
                self.pages.values(),
                key=lambda p: p.index
            )
            return sorted_pages

    async def _process_url(self, session: aiohttp.ClientSession, url: str) -> None:
        if url in self.visited_urls:
            return
        print(f"Processing {url}")
        self.visited_urls.add(url)

        # Fetch page content
        html_content = await self._fetch_page(session, url)
        if not html_content:
            return

        # Extract page content
        page = await GitbookWebFetcherService.extract_page_content(url, html_content)
        if not page:
            return

        # Check for duplicate content
        if page.content_hash in self.content_hashes:
            return

        # Set page index
        page.index = len(self.pages)

        # Add page to collection
        self.pages[url] = page
        self.content_hashes[page.content_hash] = url

        # Extract navigation links from this page
        nav_links = await GitbookWebFetcherService.gitbook_extract_navigation(self.base_url, html_content)

        #Process the discovered links
        for link in nav_links:
            if link not in self.visited_urls:
                # Add delay between requests
                await asyncio.sleep(self.delay)
                # Process the URL
                await self._process_url(session, link)

    @staticmethod
    async def extract_page_content(url: str, html_content: str) -> Optional[TextPage]:
        try:
            soup_processor = SoupProcessor(html_content)

            title = soup_processor.extract_title()
            if not title:
                title = urlparse(url).path.split('/')[-1] or "Index"
                title = title.replace('-', ' ').replace('_', ' ').title()

            last_updated = soup_processor.extract_last_updated_refs_from_soup()

            body_tag = soup_processor.find_body_content()
            if body_tag is None:
                return None
            soup_processor.clean_template_usage(body_tag)
            chapters = soup_processor.extract_chapters(content=body_tag)

            return TextPage(
                url=url,
                title=title,
                content=html_content,
                last_updated=last_updated,
                chapters=chapters,
            )
        except Exception as e:
            return None

    @staticmethod
    async def gitbook_extract_navigation(base_url: str, html_content: str) -> List[str]:
        """Extract navigation links from a page"""
        try:

            soup_processor = SoupProcessor(html_content)

            nav_links = []
            processed_urls = set()

            # Extract links from modern layout
            nav_links.extend(soup_processor.gitbook_extract_modern_nav(base_url, processed_urls))

            # Extract links from traditional layout
            nav_links.extend(soup_processor.gitbook_extract_traditional_nav(base_url, processed_urls))

            # Extract links from pagination elements
            nav_links.extend(soup_processor.gitbook_extract_pagination_links(base_url, processed_urls))

            # Extract links from search for specific class patterns
            nav_links.extend(soup_processor.gitbook_extract_class_based_nav(base_url, processed_urls))

            # Remove duplicates while preserving order
            return list(dict.fromkeys(nav_links))

        except Exception as e:
            return []

    async def _fetch_page(self, session: aiohttp.ClientSession, url: str) -> Optional[str]:
        """Fetch a page with retry logic"""
        retry_count = 0
        current_delay = self.retry_delay

        while retry_count < self.max_retries:
            try:
                async with session.get(url) as response:
                    if response.status == 429:  # Rate limit
                        retry_after = response.headers.get('Retry-After', '60')
                        wait_time = int(retry_after)

                        await asyncio.sleep(wait_time)
                        retry_count += 1
                        continue

                    if response.status == 200:
                        return await response.text()
                    else:
                        return None

            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                if retry_count < self.max_retries - 1:
                    await asyncio.sleep(current_delay)
                    current_delay *= 2  # Exponential backoff
                    retry_count += 1
                else:
                    return None
        return None




def clean_llm_response(response_text: str) -> str:
    clean_deepseek = _strip_think_tag_from_deepseek_response(_replace_role_tags_from_response(response_text))
    clean_tags = _replace_role_tags_from_response(clean_deepseek)
    return clean_tags

def _strip_think_tag_from_deepseek_response(response_text: str) -> str:
    """
    Removes any content enclosed within the "<think>" tags from the response text,
    """
    think_start = response_text.find("<think>")
    think_end = response_text.find("</think>")
    if think_start != -1 and think_end != -1:
        return response_text[think_end + 8:].strip()
    return response_text

def _replace_role_tags_from_response(response_text: str) -> str:
    return (response_text.replace("<|assistant|>", "")
            .replace("<|user|>", "")
            .replace("<|system|>", ""))



from typing import Any, List, Optional

from ..domain.priorityqueue.priority_queue_repository import PriorityQueueRepository
from ..domain.priorityqueue.queue_item import QueueItem


class PriorityQueueService:
    """Application service to manage priority queue operations"""

    def __init__(self, repository: PriorityQueueRepository):
        self.repository = repository

    def add_item(self, queue_name: str, content: Any, priority: int, metadata: Optional[dict] = None) -> QueueItem:
        """Add an item to the specified queue"""
        item = QueueItem.create(content, priority, metadata)
        self.repository.add_item(queue_name, item)
        return item

    def get_next_item(self, queue_name: str) -> Optional[QueueItem]:
        """Get and remove the highest priority item from the queue"""
        return self.repository.get_highest_priority_item(queue_name)

    def get_items(self, queue_name: str, limit: int = 10) -> List[QueueItem]:
        """Get multiple items from the queue in priority order without removing them"""
        return self.repository.get_items(queue_name, limit)

    def get_queue_length(self, queue_name: str) -> int:
        """Get the number of items in the queue"""
        return self.repository.queue_length(queue_name)

    def get_available_queues(self) -> List[str]:
        """Get a list of all available queue names"""
        return self.repository.get_queue_names()



import json


def parse_json_data(json_string: str) -> dict |None:
    try:
        # Check for empty string
        if not json_string or not json_string.strip():
            print("Empty JSON string provided")
            return {}

        # First try to extract JSON from Markdown code blocks
        import re
        match = re.search(r'```(?:json)?\s*\n(.*?)\n```', json_string, re.DOTALL)
        if match:
            json_string = match.group(1).strip()
        else:
            # If no code block found, try the existing logic
            if json_string.strip().startswith('json'):
                json_string = json_string.strip()[4:].strip()

        # Parse the JSON data
        data = json.loads(json_string)

        return data
    except json.JSONDecodeError as e:
        print(f"Error parsing JSON data: {e}")
        return {}
    except Exception as e:
        print(f"Unexpected error while parsing JSON data: {e}")
        return {}


import re
from urllib.parse import urlparse

from ...domain.web_docs.ecosystem_pattern_repository import EcosystemPatternRepository


class EcosystemLinkScorerService:
    """Service for scoring potential ecosystem-related links and content"""

    def __init__(self, pattern_repo: EcosystemPatternRepository):
        self.pattern_repo = pattern_repo

    def score(self, full_url: str, link_text: str) -> float:
        """
        Score a link based on how likely it is to be ecosystem-related
        Returns a value between 0.0 and 1.0
        """
        score = 0.0
        max_score = 3.0  # Maximum possible score

        # Parse the URL
        parsed_url = urlparse(full_url)
        domain = parsed_url.netloc
        path = parsed_url.path

        # Check domain patterns
        for eco_domain in self.pattern_repo.ecosystem_domains:
            if eco_domain in domain:
                score += 1.0
                break

        # Check path patterns
        for path_pattern in self.pattern_repo.ecosystem_path_patterns:
            if re.search(path_pattern, path):
                score += 1.0
                break

        # Check link text patterns
        for text_pattern in self.pattern_repo.ecosystem_text_patterns:
            if re.search(text_pattern, link_text):
                score += 1.0
                break

        # Normalize score to 0.0-1.0 range
        return min(score / max_score, 1.0)

    def score_page(self, page_url: str, page_title: str, page_content: str) -> float:
        """
        Score an entire page based on how likely it is to contain ecosystem information
        Returns a value between 0.0 and 1.0

        Args:
            page_url: The URL of the page
            page_title: The title of the page
            page_content: The full text content of the page
        """
        # Start with the URL and title scoring
        url_score = self.score(page_url, page_title)

        # Content-based scoring
        content_score = 0.0
        max_content_score = 2.0

        # Check content patterns
        content_matches = 0
        for content_pattern in self.pattern_repo.ecosystem_content_patterns:
            if re.search(content_pattern, page_content):
                content_matches += 1

        # Score based on number of content matches
        if content_matches >= 3:
            content_score += 1.0
        elif content_matches > 0:
            content_score += 0.5

        # Check for header patterns
        for header_pattern in self.pattern_repo.ecosystem_header_patterns:
            if re.search(header_pattern, page_content):
                content_score += 1.0
                break

        # Combined score with higher weight on content
        return min((url_score + (content_score / max_content_score) * 2) / 3, 1.0)



import re
from urllib.parse import urlparse

from ...domain.web_docs.documentation_pattern_repository import DocumentationPatternRepository


class DocLinkScorerService:
    """Service for scoring potential documentation links"""

    def __init__(self, pattern_repo: DocumentationPatternRepository):
        self.pattern_repo = pattern_repo

    def score(self, full_url: str, link_text: str) -> float:
        """
        Score a link based on how likely it is to be documentation
        Returns a value between 0.0 and 1.0
        """
        score = 0.0
        max_score = 3.0  # Maximum possible score

        # Parse the URL
        parsed_url = urlparse(full_url)
        domain = parsed_url.netloc
        path = parsed_url.path

        # Check domain patterns
        for doc_domain in self.pattern_repo.doc_domains:
            if doc_domain in domain:
                score += 1.0
                break

        # Check path patterns
        for path_pattern in self.pattern_repo.doc_path_patterns:
            if re.search(path_pattern, path):
                score += 1.0
                break

        # Check link text patterns
        for text_pattern in self.pattern_repo.doc_text_patterns:
            if re.search(text_pattern, link_text):
                score += 1.0
                break

        # Normalize score to 0.0-1.0 range
        return min(score / max_score, 1.0)